function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var Joi = _interopDefault(require('joi'));

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

class OpenAPISpecification {
  constructor({
    openapi = '3.0.0',
    info = {
      title: '',
      version: ''
    },
    components,
    paths = {},
    security = [],
    tags = [],
    servers = [],
    externalDocs
  } = {
    openapi: '3.0.0',
    info: {
      title: '',
      version: ''
    },
    paths: {}
  }) {
    this.spec = {
      openapi,
      info,
      components: _extends({
        schemas: {},
        responses: {},
        parameters: {},
        examples: {},
        requestBodies: {},
        headers: {},
        securitySchemes: {},
        links: {},
        callbacks: {}
      }, components),
      paths,
      security,
      tags,
      servers,
      externalDocs
    };
    this.setOpenApiVersion(openapi);
  }

  static isValidOpenApiVersion(version = '') {
    const match = /(\d+)\.(\d+).(\d+)/.exec(version);

    if (match) {
      const major = Number(match[1]);

      if (major >= 3) {
        return true;
      }
    }

    return false;
  }

  assertNonNull(value, name) {
    if (!value) {
      throw new Error(`${name} is not set!`);
    }
  }

  get openapi() {
    return this.spec.openapi;
  }

  setOpenApiVersion(version) {
    if (!OpenAPISpecification.isValidOpenApiVersion(version)) {
      throw new Error(`Invalid OpenAPI version: ${version}, expected format: 3.y.z`);
    }

    this.spec.openapi = version;
    return this;
  }

  get info() {
    return this.spec.info;
  }

  setInfo(info) {
    this.spec.info = info;
    return this;
  }

  get components() {
    return this.spec.components;
  }

  setComponent(componentType, name, component) {
    this.spec.components[componentType] = _extends({}, this.spec.components[componentType], {
      [name]: component
    });
    return this;
  }

  get paths() {
    return this.spec.paths;
  }

  getPathItem(path) {
    var _this$paths$path;

    return (_this$paths$path = this.paths[path]) != null ? _this$paths$path : null;
  }

  setPathItem(path, pathItem) {
    this.spec.paths[path] = pathItem;
    return this;
  }

  addPathItemParameter(path, parameter) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);

    if (!pathItem.parameters) {
      pathItem.parameters = [];
    }

    pathItem.parameters.push(parameter);
    this.spec.paths[path] = pathItem;
    return this;
  }

  addPathItemServer(path, server) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);

    if (!pathItem.servers) {
      pathItem.servers = [];
    }

    pathItem.servers.push(server);
    this.spec.paths[path] = pathItem;
    return this;
  }

  getPathItemOperation(path, method) {
    var _this$paths$path$meth, _this$paths$path2;

    return (_this$paths$path$meth = (_this$paths$path2 = this.paths[path]) === null || _this$paths$path2 === void 0 ? void 0 : _this$paths$path2[method]) != null ? _this$paths$path$meth : null;
  }

  setPathItemOperation(path, method, operation) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);
    this.spec.paths[path] = _extends({}, pathItem, {
      [method]: operation
    });
    return this;
  }

  setPathItemOperationRequestBody(path, method, requestBody) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.requestBody = requestBody;
    return this;
  }

  getPathItemOperationResponse(path, method, httpStatusCode) {
    var _this$paths$path$meth2, _this$paths$path3, _this$paths$path3$met, _this$paths$path3$met2;

    return (_this$paths$path$meth2 = (_this$paths$path3 = this.paths[path]) === null || _this$paths$path3 === void 0 ? void 0 : (_this$paths$path3$met = _this$paths$path3[method]) === null || _this$paths$path3$met === void 0 ? void 0 : (_this$paths$path3$met2 = _this$paths$path3$met.responses) === null || _this$paths$path3$met2 === void 0 ? void 0 : _this$paths$path3$met2[httpStatusCode]) != null ? _this$paths$path$meth2 : null;
  }

  setPathItemOperationResponse(path, method, httpStatusCode, response) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.responses[httpStatusCode] = response;
    return this.setPathItemOperation(path, method, operation);
  }

  addPathItemOperationTag(path, method, tag) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.tags = operation.tags ? operation.tags.concat(tag) : [tag];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationParameter(path, method, parameter) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.parameters = operation.parameters ? operation.parameters.concat(parameter) : [parameter];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationSecurityRequirement(path, method, securityRequirement) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.security = operation.security ? operation.security.concat(securityRequirement) : [securityRequirement];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationServer(path, method, server) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.servers = operation.servers ? operation.servers.concat(server) : [server];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  get security() {
    return this.spec.security;
  }

  addSecurity(security) {
    this.spec.security.push(security);
    return this;
  }

  get tags() {
    return this.spec.tags;
  }

  addTag(tag) {
    this.spec.tags.push(tag);
    return this;
  }

  get servers() {
    return this.spec.servers;
  }

  addServer(server) {
    this.spec.servers.push(server);
    return this;
  }

  get externalDocs() {
    return this.spec.externalDocs;
  }

  setExternalDocs(externalDocs) {
    this.spec.externalDocs = externalDocs;
    return this;
  }

  toJSON() {
    return this.spec;
  }

  toString(replacer, space) {
    return JSON.stringify(this.spec, replacer, space);
  }

}

const regexPattern = {
  param: /\(\?:\(\[\^\\\/]\+\?\)\)/,
  path: /^\^\\\/(?:(:?[\w\\.-]*(?:\\\/:?[\w\\.-]*)*)|(\(\?:\(\[\^\\\/]\+\?\)\)))\\/,
  rootPath: /^\^\\\/\?\(\?=\\\/\|\$\)$/
};

const getPathString = (pathRegex, keys) => {
  const pathRegexpSource = pathRegex.source;
  let regexString = pathRegexpSource;
  let regexStringExecArray = regexPattern.path.exec(regexString);
  let paramIndex = 0;

  while (regexPattern.param.test(regexString)) {
    const param = `:${keys[paramIndex].name}`;
    regexString = regexString.replace(regexPattern.param, param);
    paramIndex++;
  }

  if (regexString !== pathRegexpSource) {
    regexStringExecArray = regexPattern.path.exec(regexString);
  }

  const pathString = regexStringExecArray[1].replace(/\\\//g, '/');
  return pathString;
};

const processRoute = (spec, route, basePath = '', plugins) => {
  const routePaths = Array.isArray(route.path) ? route.path : [route.path];
  routePaths.forEach(routePath => {
    const path = `${basePath ? `${basePath}${routePath === '/' ? '' : routePath}` : routePath}`.replace(/:([a-zA-Z0-9_]+)/g, '{$1}');
    const methods = Object.keys(route.methods).filter(method => method !== '_all');

    if (!spec.paths[path]) {
      spec.setPathItem(path, {});
    }

    for (const method of methods) {
      spec.setPathItemOperation(path, method, {
        responses: {
          default: {
            description: ''
          }
        }
      });

      for (const plugin of plugins) {
        const stashedValue = plugin.stash.find(route, method);

        if (stashedValue) {
          plugin.processRoute(plugin.specification, stashedValue, {
            path,
            method
          });
        }
      }
    }
  });
};

const processExpressRouters = (spec, app, basePath = '', plugins) => {
  var _app$stack, _app$_router;

  const stack = (_app$stack = app.stack) != null ? _app$stack : (_app$_router = app._router) === null || _app$_router === void 0 ? void 0 : _app$_router.stack;

  if (!stack) {
    spec.setPathItem(basePath, {});
    return;
  }

  for (const layer of stack) {
    // terminal route
    if (layer.route) {
      processRoute(spec, layer.route, basePath, plugins);
      continue;
    }

    if (layer.name !== "router"
    /* ROUTER */
    && layer.name !== "bound dispatch"
    /* BOUND_DISPATCH */
    && layer.name !== "mounted_app"
    /* MOUNTED_APP */
    ) {
        continue;
      }

    if (regexPattern.path.test(layer.regexp.source)) {
      const parsedPath = getPathString(layer.regexp, layer.keys);
      processExpressRouters(spec, layer.handle, basePath + '/' + parsedPath, plugins);
      continue;
    }

    if (!layer.path && layer.regexp && !regexPattern.rootPath.test(layer.regexp.source)) {
      const regexPath = ' RegExp(' + layer.regexp + ') ';
      processExpressRouters(spec, layer.handle, basePath + '/' + regexPath, plugins);
      continue;
    }

    processExpressRouters(spec, layer.handle, basePath, plugins);
  }
};

class Stash {
  constructor(stashKey) {
    this.find = (route, method) => {
      for (const layer of route.stack) {
        if (layer.method === method) {
          const descriptor = Object.getOwnPropertyDescriptor(layer.handle, this.stashKey);

          if (descriptor) {
            return descriptor.value;
          }
        }
      }

      return null;
    };

    this.store = (handler, value) => {
      Object.defineProperty(handler, this.stashKey, {
        value,
        enumerable: true
      });
    };

    this.stashKey = stashKey;
  }

}

const isSchemaType = (schema, type) => {
  return schema.type === type || schema.$_terms.metas.some(meta => meta.baseType === type);
};
const isRuleName = (rule, name) => {
  return rule.name === name;
};
const isDeprecated = schema => {
  return schema.$_terms.metas.some(meta => meta.deprecated === true);
};
const extractAlternativeSchemas = conditionals => {
  const alternativeSchemas = [];

  for (const conditional of conditionals) {
    if (conditional.ref) {
      if (conditional.then) {
        alternativeSchemas.push(conditional.then);
      }

      if (conditional.otherwise) {
        alternativeSchemas.push(conditional.otherwise);
      }

      if (conditional.switch) {
        for (const switchCase of conditional.switch) {
          alternativeSchemas.push(switchCase.then);

          if (switchCase.otherwise) {
            alternativeSchemas.push(switchCase.otherwise);
          }
        }
      }
    } else if ('schema' in conditional && conditional.schema) {
      alternativeSchemas.push(conditional.schema);
    }
  }

  return alternativeSchemas;
};
const getMetas = schema => {
  return schema.$_terms.metas.reduce((metas, meta) => {
    if (typeof meta === 'object' && !Array.isArray(meta)) {
      Object.assign(metas, meta);
    }

    return metas;
  }, {});
};

function parseAnySchema(schema) {
  const result = {
    schema: {},
    meta: {}
  };
  const metas = getMetas(schema);

  if (metas.customType === 'file') {
    result.schema.type = 'string';
    result.schema.format = 'binary';
  }

  return result;
}

function parseAlternativesSchema(schema) {
  var _flags$match;

  const result = {
    schema: {},
    meta: {}
  };
  const _flags = schema._flags;
  const matchMode = `${(_flags$match = _flags.match) != null ? _flags$match : 'any'}Of`;
  result.schema[matchMode] = [];
  const $_terms = schema.$_terms;
  const alternativeSchemas = extractAlternativeSchemas($_terms.matches);

  for (const alternativeSchema of alternativeSchemas) {
    const matchResult = parseJoiSchema(alternativeSchema);
    result.schema[matchMode].push(matchResult.schema);
  }

  return result;
}

function parseArraySchema(schema) {
  const result = {
    schema: {
      type: 'array',
      items: {}
    },
    meta: {}
  };
  const $_terms = schema.$_terms;
  const itemResults = $_terms.items.map(itemSchema => parseJoiSchema(itemSchema));

  if (itemResults.length === 1) {
    result.schema.items = itemResults[0].schema;
  } else if (itemResults.length > 1) {
    result.schema.items = {
      anyOf: itemResults.map(({
        schema
      }) => schema)
    };
  }

  for (const rule of schema._rules) {
    if (isRuleName(rule, 'length')) {
      result.schema.minItems = rule.args.limit;
      result.schema.maxItems = rule.args.limit;
    }

    if (isRuleName(rule, 'min')) {
      result.schema.minItems = rule.args.limit;
    }

    if (isRuleName(rule, 'max')) {
      result.schema.maxItems = rule.args.limit;
    }

    if (isRuleName(rule, 'unique')) {
      result.schema.uniqueItems = true;
    }
  }

  return result;
}

function parseBooleanSchema(schema) {
  var _result$schema$oneOf3;

  const result = {
    schema: {
      oneOf: [{
        type: 'boolean'
      }]
    },
    meta: {}
  };
  const $_terms = schema.$_terms;
  const stringEnum = [];
  const integerEnum = [];
  const keys = ['falsy', 'truthy'];

  for (const key of keys) {
    const term = $_terms[key];

    if (!term) {
      continue;
    }

    for (const value of term._values.values()) {
      if (typeof value === 'number') {
        integerEnum.push(value);
      } else if (typeof value === 'string') {
        stringEnum.push(value);
      }
    }
  }

  if (stringEnum.length > 0) {
    var _result$schema$oneOf;

    (_result$schema$oneOf = result.schema.oneOf) === null || _result$schema$oneOf === void 0 ? void 0 : _result$schema$oneOf.push({
      type: 'string',
      enum: stringEnum
    });
  }

  if (integerEnum.length > 0) {
    var _result$schema$oneOf2;

    (_result$schema$oneOf2 = result.schema.oneOf) === null || _result$schema$oneOf2 === void 0 ? void 0 : _result$schema$oneOf2.push({
      type: 'integer',
      enum: integerEnum
    });
  }

  if (((_result$schema$oneOf3 = result.schema.oneOf) === null || _result$schema$oneOf3 === void 0 ? void 0 : _result$schema$oneOf3.length) === 1) {
    result.schema = result.schema.oneOf[0];
  }

  return result;
}

function parseBinarySchema(schema) {
  const result = {
    schema: {
      type: 'string'
    },
    meta: {}
  };
  const _flags = schema._flags;

  if (_flags.encoding === 'base64') {
    result.schema.format = 'byte';
  } else {
    result.schema.format = 'binary';
  }

  return result;
}

function parseDateSchema(schema) {
  const result = {
    schema: {
      type: 'string'
    },
    meta: {}
  };
  const _flags = schema._flags;

  if (_flags.format === 'iso') {
    result.schema.format = 'date-time';
  }

  if (_flags.format === 'javascript' || _flags.format === 'unix') {
    result.schema.type = 'integer';
  }

  return result;
}

function parseNumberSchema(schema) {
  const result = {
    schema: {},
    meta: {}
  };

  if (schema._singleRules.get('integer')) {
    result.schema.type = 'integer';
  } else {
    result.schema.type = 'number';

    const precisionRule = schema._singleRules.get('precision');

    if (precisionRule) {
      result.schema.format = 'double';
    } else {
      result.schema.format = 'float';
    }
  }

  for (const rule of schema._rules) {
    if (isRuleName(rule, 'greater')) {
      result.schema.minimum = rule.args.limit;
      result.schema.exclusiveMinimum = true;
    }

    if (isRuleName(rule, 'less')) {
      result.schema.maximum = rule.args.limit;
      result.schema.exclusiveMaximum = true;
    }

    if (isRuleName(rule, 'min')) {
      result.schema.minimum = rule.args.limit;
    }

    if (isRuleName(rule, 'max')) {
      result.schema.maximum = rule.args.limit;
    }

    if (isRuleName(rule, 'port')) {
      result.schema.minimum = 0;
      result.schema.maximum = 65535;
    }

    if (isRuleName(rule, 'multiple')) {
      result.schema.multipleOf = rule.args.base;
    }
  }

  const signRule = schema._singleRules.get('sign');

  if (signRule) {
    if (signRule.args.sign === 'negative') {
      var _result$schema$maximu;

      const maximum = (_result$schema$maximu = result.schema.maximum) != null ? _result$schema$maximu : -1;
      result.schema.maximum = maximum < 0 ? maximum : -1;
    }

    if (signRule.args.sign === 'positive') {
      var _result$schema$minimu;

      const minimum = (_result$schema$minimu = result.schema.minimum) != null ? _result$schema$minimu : 0;
      result.schema.minimum = minimum >= 0 ? minimum : 0;
    }
  }

  return result;
}

function parseObjectSchema(schema) {
  var _$_terms$keys;

  const result = {
    schema: {
      type: 'object',
      properties: {}
    },
    meta: {}
  };
  const _flags = schema._flags;

  if (_flags.unknown !== true) {
    result.schema.additionalProperties = false;
  }

  const $_terms = schema.$_terms;
  const required = [];
  const keys = (_$_terms$keys = $_terms.keys) != null ? _$_terms$keys : [];

  for (const key of keys) {
    const name = key.key;
    const keyResult = parseJoiSchema(key.schema);
    result.schema.properties[name] = keyResult.schema;

    if (keyResult.meta.required) {
      required.push(name);
    }
  }

  for (const rule of schema._rules) {
    if (isRuleName(rule, 'length')) {
      result.schema.minProperties = rule.args.limit;
      result.schema.maxProperties = rule.args.limit;
    }

    if (isRuleName(rule, 'min')) {
      result.schema.minProperties = rule.args.limit;
    }

    if (isRuleName(rule, 'max')) {
      result.schema.maxProperties = rule.args.limit;
    }
  }

  if (required.length) {
    result.schema.required = required;
  }

  return result;
}

function parseStringSchema(schema) {
  const result = {
    schema: {
      type: 'string'
    },
    meta: {}
  };

  for (const rule of schema._rules) {
    if (isRuleName(rule, 'base64')) {
      result.schema.format = 'byte';
    }

    if (isRuleName(rule, 'email')) {
      result.schema.format = 'email';
    }

    if (isRuleName(rule, 'guid')) {
      result.schema.format = 'uuid';
    }

    if (isRuleName(rule, 'ip')) {
      var _rule$args$options$ve;

      result.schema.format = ((_rule$args$options$ve = rule.args.options.version) === null || _rule$args$options$ve === void 0 ? void 0 : _rule$args$options$ve.length) === 1 ? rule.args.options.version[0] : 'ip';
    }

    if (isRuleName(rule, 'isoDate')) {
      result.schema.format = 'date-time';
    }

    if (isRuleName(rule, 'uri')) {
      result.schema.format = 'uri';
    }

    if (isRuleName(rule, 'alphanum')) {
      var _schema$_preferences;

      const isStrict = ((_schema$_preferences = schema._preferences) === null || _schema$_preferences === void 0 ? void 0 : _schema$_preferences.convert) === false;

      const caseRule = schema._singleRules.get('case');

      if (caseRule && !isStrict) {
        if (caseRule.args.direction === 'upper') {
          result.schema.pattern = /^[A-Z0-9]*$/.source;
        }

        if (caseRule.args.direction === 'lower') {
          result.schema.pattern = /^[a-z0-9]*$/.source;
        }
      } else {
        result.schema.pattern = /^[a-zA-Z0-9]*$/.source;
      }
    }

    if (isRuleName(rule, 'token')) {
      result.schema.pattern = /^[a-zA-Z0-9_]*$/.source;
    }

    if (isRuleName(rule, 'pattern')) {
      var _rule$args$options;

      if (typeof rule.args.options !== 'string' && ((_rule$args$options = rule.args.options) === null || _rule$args$options === void 0 ? void 0 : _rule$args$options.invert)) {
        result.schema.not = _extends({}, result.schema.not, {
          pattern: rule.args.regex.source
        });
      } else {
        result.schema.pattern = rule.args.regex.source;
      }
    } else if (rule.regex) {
      result.schema.pattern = rule.regex.source;
    }

    if (isRuleName(rule, 'length')) {
      result.schema.minLength = rule.args.limit;
      result.schema.maxLength = rule.args.limit;
    }

    if (isRuleName(rule, 'min')) {
      result.schema.minLength = rule.args.limit;
    }

    if (isRuleName(rule, 'max')) {
      result.schema.minLength = rule.args.limit;
    }
  }

  return result;
}

function parseJoiSchema(schema) {
  const result = {
    schema: {},
    meta: {}
  };
  const _flags = schema._flags;

  if (_flags.presence === 'forbidden') {
    return result;
  }

  if (_flags.presence === 'required') {
    result.meta.required = true;
  }

  if (_flags.presence === 'optional') {
    result.meta.required = false;
  }

  if (_flags.description) {
    result.schema.description = _flags.description;
  }

  if (_flags.label) {
    result.schema.title = _flags.label;
  }

  if (_flags.default) {
    if (typeof _flags.default !== 'function') {
      result.schema.default = _flags.default;
    }
  }

  if (isDeprecated(schema)) {
    result.schema.deprecated = true;
  }

  if (schema._valids) {
    if (schema._valids._values.has(null)) {
      result.schema.nullable = true;
    }

    if (_flags.only) {
      result.schema.enum = Array.from(schema._valids._values).filter(value => value !== null);
    }
  } else if (schema._invalids) {
    result.schema.not = {
      enum: Array.from(schema._invalids._values)
    };
  }

  const $_terms = schema.$_terms;

  if ($_terms.examples) {
    if ($_terms.examples.length === 1) {
      result.schema.example = $_terms.examples[0];
    } else {
      result.schema.examples = $_terms.examples;
    }
  }

  if ($_terms.whens) {
    result.schema.anyOf = [];
    const alternativeSchemas = extractAlternativeSchemas($_terms.whens);

    for (const alternativeSchema of alternativeSchemas) {
      const matchResult = parseJoiSchema(alternativeSchema);
      result.schema.anyOf.push(matchResult.schema);
    }
  }

  let resultOverride = {
    schema: {},
    meta: {}
  };

  if (isSchemaType(schema, 'array')) {
    resultOverride = parseArraySchema(schema);
  }

  if (isSchemaType(schema, 'binary')) {
    resultOverride = parseBinarySchema(schema);
  }

  if (isSchemaType(schema, 'boolean')) {
    resultOverride = parseBooleanSchema(schema);
  }

  if (isSchemaType(schema, 'date')) {
    resultOverride = parseDateSchema(schema);
  }

  if (isSchemaType(schema, 'number')) {
    resultOverride = parseNumberSchema(schema);
  }

  if (isSchemaType(schema, 'object')) {
    resultOverride = parseObjectSchema(schema);
  }

  if (isSchemaType(schema, 'string')) {
    resultOverride = parseStringSchema(schema);
  }

  if (isSchemaType(schema, 'any')) {
    resultOverride = parseAnySchema(schema);
  }

  if (isSchemaType(schema, 'alternatives')) {
    resultOverride = parseAlternativesSchema(schema);
  }

  Object.assign(result.schema, resultOverride.schema);
  Object.assign(result.meta, resultOverride.meta);
  return result;
}

function _catch(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }

  if (result && result.then) {
    return result.then(void 0, recover);
  }

  return result;
}

class RequestValidationError extends Error {
  constructor(validationError, segment) {
    super(validationError.message);
    this.segment = segment;
    this.validationError = validationError;
  }

}
const parameterLocationBySegment = {
  cookies: 'cookie',
  signedCookies: 'cookie',
  headers: 'header',
  params: 'path',
  query: 'query'
};
const defaultRequestSegmentOrder = ['headers', 'params', 'query', 'cookies', 'signedCookies', 'body'];
const getJoiRequestValidatorPlugin = ({
  segmentOrder = defaultRequestSegmentOrder
} = {}) => {
  const joiRequestValidatorPlugin = {
    name: 'joi-request-validator',
    getMiddleware: (internals, joiRequestSchema, joiValidationOptions = {}) => {
      const schemaMap = new Map();

      for (const segment of segmentOrder) {
        const schema = joiRequestSchema[segment];

        if (schema) {
          schemaMap.set(segment, Joi.compile(schema));
        }
      }

      const validationOptions = _extends({}, joiValidationOptions);

      const extendedValidationResult = validationOptions.debug || validationOptions.warnings;

      const requestValidationMiddleware = function (req, _res, next) {
        try {
          const _temp = _catch(function () {
            return Promise.resolve(segmentOrder.reduce((promise, segment) => {
              return promise.then(() => {
                const schema = schemaMap.get(segment);

                if (!schema) {
                  return null;
                }

                return schema.validateAsync(req[segment], validationOptions).then(result => {
                  if (extendedValidationResult) {
                    req[segment] = result.value;
                  } else {
                    req[segment] = result;
                  }

                  return null;
                }).catch(error => {
                  throw new RequestValidationError(error, segment);
                });
              });
            }, Promise.resolve(null))).then(function () {
              next();
            });
          }, function (error) {
            next(error);
          });

          return Promise.resolve(_temp && _temp.then ? _temp.then(function () {}) : void 0);
        } catch (e) {
          return Promise.reject(e);
        }
      };

      internals.stash.store(requestValidationMiddleware, schemaMap);
      return requestValidationMiddleware;
    },
    processRoute: (specification, schemaMap, {
      path,
      method
    }) => {
      for (const [segment, schema] of schemaMap.entries()) {
        if (!schema) {
          continue;
        }

        if (segment === 'body') {
          const result = parseJoiSchema(schema);
          const requestBody = {
            content: {
              'application/json': {
                schema: result.schema
              }
            }
          };
          specification.setPathItemOperationRequestBody(path, method, requestBody);
          return;
        }

        const location = parameterLocationBySegment[segment];
        const result = parseJoiSchema(schema);
        const {
          properties,
          required = []
        } = result.schema;

        if (properties) {
          for (const [name, schema] of Object.entries(properties)) {
            const parameterObject = {
              name,
              in: location,
              schema: schema,
              required: required.includes(name)
            };

            if ('deprecated' in schema) {
              parameterObject.deprecated = schema.deprecated;
            }

            specification.addPathItemOperationParameter(path, method, parameterObject);
          }
        }
      }
    }
  };
  return joiRequestValidatorPlugin;
};

class ResponseValidationError extends Error {
  constructor(validationError, segment) {
    super(validationError.message);
    this.segment = segment;
    this.validationError = validationError;
  }

}
const defaultResponseSegmentOrder = ['body', 'headers'];
const getJoiResponseValidatorPlugin = ({
  skipValidation = false,
  segmentOrder = defaultResponseSegmentOrder
} = {}) => {
  const joiResponseValidatorPlugin = {
    name: 'joi-response-validator',
    getMiddleware: (internals, joiResponseSchema, joiValidationOptions = {}) => {
      const schemaMap = new Map();

      for (const [code, schemaBySegment] of Object.entries(joiResponseSchema)) {
        schemaMap.set(code, Object.entries(schemaBySegment).reduce((bySegment, [segment, schema]) => {
          if (schema) {
            bySegment[segment] = Joi.compile(schema);
          }

          return bySegment;
        }, {}));
      }

      const validationOptions = _extends({}, joiValidationOptions);

      const validationMiddleware = function (_req, res, next) {
        try {
          if (skipValidation) {
            return Promise.resolve(next());
          }

          const originalSend = res.send;

          res.send = function validateAndSendResponse(...args) {
            const body = args[0];
            const isJsonContent = /application\/json/.test(String(res.get('content-type')));
            const value = {
              body: isJsonContent ? JSON.parse(body) : body,
              headers: res.getHeaders()
            };
            const schemaBySegment = schemaMap.has(String(res.statusCode)) ? schemaMap.get(String(res.statusCode)) : schemaMap.get('default');

            if (!schemaBySegment) {
              throw new Error(`Validation Schema not found for Response(${res.statusCode})`);
            }

            for (const segment of segmentOrder) {
              const schema = schemaBySegment[segment];

              if (!schema) {
                continue;
              }

              const validationResult = schema.validate(value[segment], validationOptions);

              if (validationResult.error) {
                throw new ResponseValidationError(validationResult.error, segment);
              }

              if (segment === 'body') {
                value[segment] = validationResult.value;
              }
            }

            return originalSend.apply(res, [isJsonContent ? JSON.stringify(value.body) : value.body]);
          };

          next();
          return Promise.resolve();
        } catch (e) {
          return Promise.reject(e);
        }
      };

      internals.stash.store(validationMiddleware, schemaMap);
      return validationMiddleware;
    },
    processRoute: (specification, schemaMap, {
      path,
      method
    }) => {
      for (const [key, schemaBySegment] of schemaMap.entries()) {
        if (!schemaBySegment) {
          continue;
        }

        const httpStatusCode = key;

        if (schemaBySegment.body) {
          const result = parseJoiSchema(schemaBySegment.body);

          const responseObject = _extends({}, specification.getPathItemOperationResponse(path, method, httpStatusCode), {
            description: '',
            content: {
              'application/json': {
                schema: result.schema
              }
            }
          });

          specification.setPathItemOperationResponse(path, method, httpStatusCode, responseObject);
        }

        if (schemaBySegment.headers) {
          const result = parseJoiSchema(schemaBySegment.headers);
          const {
            properties,
            required = []
          } = result.schema;
          const headersObject = {};

          if (properties) {
            for (const [name, schema] of Object.entries(properties)) {
              const headerObject = {
                schema: schema,
                required: required.includes(name)
              };

              if ('deprecated' in schema) {
                headerObject.deprecated = schema.deprecated;
              }

              headersObject[name] = headerObject;
            }
          }

          const responseObject = _extends({}, specification.getPathItemOperationResponse(path, method, httpStatusCode), {
            description: '',
            headers: headersObject
          });

          specification.setPathItemOperationResponse(path, method, httpStatusCode, responseObject);
        }
      }
    }
  };
  return joiResponseValidatorPlugin;
};

class ExpressOpenAPI {
  constructor({
    specification = new OpenAPISpecification(),
    useStringStashKey = false
  } = {}) {
    this.plugins = [];

    this.registerPlugin = plugin => {
      const stashSymbol = Symbol(plugin.name);
      const stashKey = this.useStringStashKey ? stashSymbol.toString() : stashSymbol;
      const internals = {
        specification: this.specification,
        stash: new Stash(stashKey)
      };

      const getMiddleware = (...params) => {
        return plugin.getMiddleware(internals, ...params);
      };

      const registeredPlugin = {
        name: plugin.name,
        specification: internals.specification,
        stash: internals.stash,
        processRoute: plugin.processRoute
      };
      this.plugins.push(registeredPlugin);
      return getMiddleware;
    };

    this.populateSpecification = (app, basePath) => {
      processExpressRouters(this.specification, app, basePath, this.plugins);
      return this.specification;
    };

    this.specification = specification;
    this.useStringStashKey = useStringStashKey;
  }

}

exports.ExpressOpenAPI = ExpressOpenAPI;
exports.OpenAPISpecification = OpenAPISpecification;
exports.RequestValidationError = RequestValidationError;
exports.ResponseValidationError = ResponseValidationError;
exports.getJoiRequestValidatorPlugin = getJoiRequestValidatorPlugin;
exports.getJoiResponseValidatorPlugin = getJoiResponseValidatorPlugin;
exports.parseJoiSchema = parseJoiSchema;
//# sourceMappingURL=index.js.map
